from collections.abc import Mapping
from copy import copy
from pathlib import Path
from typing import Any, Optional

import click
from click.core import ParameterSource
from typer.rich_utils import rich_format_help

from dagster_dg.cli.shared_options import (
    GLOBAL_OPTIONS,
    dg_editable_dagster_options,
    dg_global_options,
)
from dagster_dg.component import RemoteComponentRegistry, RemoteComponentType
from dagster_dg.component_key import ComponentKey
from dagster_dg.config import (
    DgRawCliConfig,
    DgRawWorkspaceConfig,
    DgWorkspaceScaffoldProjectOptions,
    get_config_from_cli_context,
    has_config_on_cli_context,
    normalize_cli_config,
    set_config_on_cli_context,
)
from dagster_dg.context import DgContext
from dagster_dg.scaffold import (
    scaffold_component_instance,
    scaffold_component_type,
    scaffold_project,
    scaffold_workspace,
)
from dagster_dg.utils import (
    DgClickCommand,
    DgClickGroup,
    exit_with_error,
    generate_missing_component_type_error_message,
    json_schema_property_to_click_option,
    not_none,
    parse_json_option,
    snakecase,
)

DEFAULT_WORKSPACE_NAME = "dagster-workspace"


@click.group(name="scaffold", cls=DgClickGroup)
def scaffold_group():
    """Commands for scaffolding Dagster code."""


# ########################
# ##### WORKSPACE
# ########################


@scaffold_group.command(name="workspace", cls=DgClickCommand)
@click.argument("name", type=str, default=DEFAULT_WORKSPACE_NAME)
@dg_editable_dagster_options
@dg_global_options
def scaffold_workspace_command(
    name: str,
    use_editable_dagster: Optional[str],
    **global_options: object,
):
    """Initialize a new Dagster workspace.

    The scaffolded workspace folder has the following structure:

    \b
    ├── dagster-workspace
    │   ├── projects
    |   |   └── <Dagster projects go here>
    |   ├── libraries
    |   |   └── <Shared packages go here>
    │   └── pyproject.toml

    """  # noqa: D301
    workspace_config = DgRawWorkspaceConfig(
        scaffold_project_options=DgWorkspaceScaffoldProjectOptions.get_raw_from_cli(
            use_editable_dagster,
        )
    )
    scaffold_workspace(name, workspace_config)


# ########################
# ##### PROJECT
# ########################


@scaffold_group.command(name="project", cls=DgClickCommand)
@click.argument("path", type=Path)
@click.option(
    "--skip-venv",
    is_flag=True,
    default=False,
    help="Do not create a virtual environment for the project.",
)
@click.option(
    "--populate-cache/--no-populate-cache",
    is_flag=True,
    default=True,
    help="Whether to automatically populate the component type cache for the project.",
    hidden=True,
)
@dg_editable_dagster_options
@dg_global_options
def scaffold_project_command(
    path: Path,
    skip_venv: bool,
    populate_cache: bool,
    use_editable_dagster: Optional[str],
    **global_options: object,
) -> None:
    """Scaffold a Dagster project file structure and a uv-managed virtual environment scoped
    to the project.

    This command can be run inside or outside of a workspace directory. If run inside a workspace,
    the project will be created within the workspace directory's project directory.

    The project file structure defines a Python package with some pre-existing internal
    structure:

    \b
    ├── <name>
    │   ├── __init__.py
    │   ├── defs
    │   ├── definitions.py
    │   └── lib
    │       └── __init__.py
    ├── <name>_tests
    │   └── __init__.py
    └── pyproject.toml

    The `<name>.lib` directory holds Python objects that can be targeted by the `dg scaffold` command or have dg-inspectable metadata. Custom component types in the project live in `<name>.lib`. These types can be created with `dg scaffold component-type`.
    """  # noqa: D301
    cli_config = normalize_cli_config(global_options, click.get_current_context())
    dg_context = DgContext.from_file_discovery_and_command_line_config(Path.cwd(), cli_config)

    abs_path = path.resolve()
    if dg_context.is_workspace:
        if dg_context.has_project(abs_path.relative_to(dg_context.workspace_root_path)):
            exit_with_error(f"The current workspace already specifies a project at {abs_path}.")
        elif abs_path.exists():
            exit_with_error(f"A file or directory already exists at {abs_path}.")

    scaffold_project(
        abs_path,
        dg_context,
        use_editable_dagster=use_editable_dagster,
        skip_venv=skip_venv,
        populate_cache=populate_cache,
    )


# ########################
# ##### COMPONENT
# ########################


# The `dg scaffold component` command is special because its subcommands are dynamically generated
# from the registered component types in the project. Because the registered component types
# depend on the built-in component library we are using, we cannot resolve them until we have the
# built-in component library, which can be set via a global option, e.g.:
#
#     dg --builtin-component-lib dagster_components.test ...
#
# To handle this, we define a custom click.Group subclass that loads the commands on demand.
class ComponentScaffoldGroup(DgClickGroup):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._commands_defined = False

    def get_command(self, cli_context: click.Context, cmd_name: str) -> Optional[click.Command]:
        if not self._commands_defined:
            self._define_commands(cli_context)
        cmd = super().get_command(cli_context, cmd_name)
        if cmd is None:
            exit_with_error(generate_missing_component_type_error_message(cmd_name))
        return cmd

    def list_commands(self, cli_context: click.Context) -> list[str]:
        if not self._commands_defined:
            self._define_commands(cli_context)
        return super().list_commands(cli_context)

    def _define_commands(self, cli_context: click.Context) -> None:
        """Dynamically define a command for each registered component type."""
        if not has_config_on_cli_context(cli_context):
            cli_context.invoke(not_none(self.callback), **cli_context.params)
        config = get_config_from_cli_context(cli_context)
        dg_context = DgContext.for_defined_registry_environment(Path.cwd(), config)

        registry = RemoteComponentRegistry.from_dg_context(dg_context)
        for key, component_type in registry.items():
            command = _create_component_scaffold_subcommand(key, component_type)
            self.add_command(command)


class ComponentScaffoldSubCommand(DgClickCommand):
    # We have to override this because the implementation of `format_help` used elsewhere will only
    # pull parameters directly off the target command. For these component scaffold subcommands  we need
    # to expose the global options, which are defined on the preceding group rather than the command
    # itself.
    def format_help(self, context: click.Context, formatter: click.HelpFormatter):
        """Customizes the help to include hierarchical usage."""
        if not isinstance(self, click.Command):
            raise ValueError("This mixin is only intended for use with click.Command instances.")

        # This is a hack. We pass the help format func a modified version of the command where the global
        # options are attached to the command itself. This will cause them to be included in the
        # help output.
        cmd_copy = copy(self)
        cmd_copy.params = [
            *cmd_copy.params,
            *(GLOBAL_OPTIONS.values()),
        ]
        rich_format_help(obj=cmd_copy, ctx=context, markup_mode="rich")

    def format_usage(self, context: click.Context, formatter: click.HelpFormatter) -> None:
        if not isinstance(self, click.Command):
            raise ValueError("This mixin is only intended for use with click.Command instances.")
        arg_pieces = self.collect_usage_pieces(context)
        command_parts = context.command_path.split(" ")
        command_parts.insert(-1, "[GLOBAL OPTIONS]")
        return formatter.write_usage(" ".join(command_parts), " ".join(arg_pieces))


# We have to override the usual Click processing of `--help` here. The issue is
# that click will process this option before processing anything else, but because we are
# dynamically generating subcommands based on the content of other options, the output of --help
# actually depends on these other options. So we opt out of Click's short-circuiting
# behavior of `--help` by setting `help_option_names=[]`, ensuring that we can process the other
# options first and generate the correct subcommands. We then add a custom `--help` option that
# gets invoked inside the callback.
@scaffold_group.group(
    name="component",
    cls=ComponentScaffoldGroup,
    invoke_without_command=True,
    context_settings={"help_option_names": []},
)
@click.option("-h", "--help", "help_", is_flag=True, help="Show this message and exit.")
@dg_global_options
@click.pass_context
def scaffold_component_group(context: click.Context, help_: bool, **global_options: object) -> None:
    """Scaffold of a Dagster component."""
    # Click attempts to resolve subcommands BEFORE it invokes this callback.
    # Therefore we need to manually invoke this callback during subcommand generation to make sure
    # it runs first. It will be invoked again later by Click. We make it idempotent to deal with
    # that.
    if not has_config_on_cli_context(context):
        cli_config = normalize_cli_config(global_options, context)
        set_config_on_cli_context(context, cli_config)
    if help_:
        click.echo(context.get_help())
        context.exit(0)


def _core_scaffold(
    cli_context: click.Context,
    cli_config: DgRawCliConfig,
    component_key: ComponentKey,
    instance_name: str,
    key_value_params,
    json_params,
) -> None:
    dg_context = DgContext.for_project_environment(Path.cwd(), cli_config)
    registry = RemoteComponentRegistry.from_dg_context(dg_context)
    if not registry.has(component_key):
        exit_with_error(f"Component type `{component_key.to_typename()}` not found.")
    elif dg_context.has_component_instance(instance_name):
        exit_with_error(f"A component instance named `{instance_name}` already exists.")

    # Specified key-value params will be passed to this function with their default value of
    # `None` even if the user did not set them. Filter down to just the ones that were set by
    # the user.
    user_provided_key_value_params = {
        k: v
        for k, v in key_value_params.items()
        if cli_context.get_parameter_source(k) == ParameterSource.COMMANDLINE
    }
    if json_params is not None and user_provided_key_value_params:
        exit_with_error(
            "Detected params passed as both --json-params and individual options. These are mutually exclusive means of passing"
            " component generation parameters. Use only one.",
        )
    elif json_params:
        scaffold_params = json_params
    elif user_provided_key_value_params:
        scaffold_params = user_provided_key_value_params
    else:
        scaffold_params = None

    scaffold_component_instance(
        Path(dg_context.defs_path) / instance_name,
        component_key.to_typename(),
        scaffold_params,
        dg_context,
    )


def _create_component_shim_scaffold_subcommand(
    component_key: ComponentKey, command_name: str
) -> DgClickCommand:
    # We need to "reset" the help option names to the default ones because we inherit the parent
    # value of context settings from the parent group, which has been customized.
    @click.command(
        name=command_name,
        cls=ComponentScaffoldSubCommand,
        context_settings={"help_option_names": ["-h", "--help"]},
    )
    @click.argument("instance_name", type=str)
    @dg_global_options
    @click.pass_context
    def scaffold_component_shim_command(
        cli_context: click.Context,
        instance_name: str,
        **global_options: object,
    ) -> None:
        """Scaffold of a definition."""
        cli_config = normalize_cli_config(global_options, cli_context)
        _core_scaffold(cli_context, cli_config, component_key, instance_name, {}, {})

    return scaffold_component_shim_command


def _create_component_scaffold_subcommand(
    component_key: ComponentKey, component_type: RemoteComponentType
) -> DgClickCommand:
    # We need to "reset" the help option names to the default ones because we inherit the parent
    # value of context settings from the parent group, which has been customized.
    @click.command(
        name=component_key.to_typename(),
        cls=ComponentScaffoldSubCommand,
        context_settings={"help_option_names": ["-h", "--help"]},
    )
    @click.argument("component_instance_name", type=str)
    @click.option(
        "--json-params",
        type=str,
        default=None,
        help="JSON string of component parameters.",
        callback=parse_json_option,
    )
    @click.pass_context
    def scaffold_component_command(
        cli_context: click.Context,
        component_instance_name: str,
        json_params: Mapping[str, Any],
        **key_value_params: Any,
    ) -> None:
        f"""Scaffold of a {component_type.name} component.

        This command must be run inside a Dagster project directory. The component scaffold will be
        placed in submodule `<project_name>.components.<COMPONENT_NAME>`.

        Components can optionally be passed scaffold parameters. There are two ways to do this:

        (1) Passing a single --json-params option with a JSON string of parameters. For example:

            dg component scaffold foo.bar my_component --json-params '{{"param1": "value", "param2": "value"}}'`.

        (2) Passing each parameter as an option. For example:

            dg component scaffold foo.bar my_component --param1 value1 --param2 value2`

        It is an error to pass both --json-params and key-value pairs as options.
        """
        cli_config = get_config_from_cli_context(cli_context)
        _core_scaffold(
            cli_context,
            cli_config,
            component_key,
            component_instance_name,
            key_value_params,
            json_params,
        )

    # If there are defined scaffold params, add them to the command
    schema = component_type.scaffold_params_schema
    if schema:
        for key, field_info in schema["properties"].items():
            # All fields are currently optional because they can also be passed under
            # `--json-params`
            option = json_schema_property_to_click_option(key, field_info, required=False)
            scaffold_component_command.params.append(option)

    return scaffold_component_command


# ########################
# ##### COMPONENT SHIMS
# ########################

SHIM_COMPONENTS = {
    ComponentKey("dagster_components.dagster", "RawAssetComponent"): "asset",
    ComponentKey("dagster_components.dagster", "RawSensorComponent"): "sensor",
    ComponentKey("dagster_components.dagster", "RawScheduleComponent"): "schedule",
}

for key, command_name in SHIM_COMPONENTS.items():
    scaffold_group.add_command(
        _create_component_shim_scaffold_subcommand(key, command_name),
    )

# ########################
# ##### COMPONENT TYPE
# ########################


@scaffold_group.command(name="component-type", cls=DgClickCommand)
@click.argument("name", type=str)
@dg_global_options
@click.pass_context
def scaffold_component_type_command(
    context: click.Context, name: str, **global_options: object
) -> None:
    """Scaffold of a custom Dagster component type.

    This command must be run inside a Dagster project directory. The component type scaffold
    will be placed in submodule `<project_name>.lib.<name>`.
    """
    cli_config = normalize_cli_config(global_options, context)
    dg_context = DgContext.for_component_library_environment(Path.cwd(), cli_config)
    registry = RemoteComponentRegistry.from_dg_context(dg_context)

    module_name = snakecase(name)
    component_key = ComponentKey(
        name=name, namespace=dg_context.default_component_library_module_name
    )
    if registry.has(component_key):
        exit_with_error(f"Component type`{component_key.to_typename()}` already exists.")

    scaffold_component_type(dg_context, name, module_name)



from dataclasses import dataclass
from typing import Literal, Optional, Union

from typing_extensions import TypeAlias

DgDefinitionMetadata: TypeAlias = Union[
    "DgAssetMetadata", "DgSensorMetadata", "DgScheduleMetadata", "DgJobMetadata"
]


@dataclass
class DgAssetMetadata:
    type: Literal["asset"]
    key: str
    group: Optional[str]
    deps: list[str]
    kinds: list[str]
    description: Optional[str]
    automation_condition: Optional[str] = None


@dataclass
class DgSensorMetadata:
    type: Literal["sensor"]
    name: str


@dataclass
class DgScheduleMetadata:
    type: Literal["schedule"]
    name: str
    cron_schedule: str


@dataclass
class DgJobMetadata:
    type: Literal["job"]
    name: str



import subprocess
import sys
from collections.abc import Mapping, Sequence
from pathlib import Path
from typing import Any

import click

from dagster_dg.check import check_yaml as check_yaml_fn
from dagster_dg.cli.dev import format_forwarded_option
from dagster_dg.cli.shared_options import dg_global_options
from dagster_dg.cli.utils import create_dagster_cli_cmd
from dagster_dg.config import normalize_cli_config
from dagster_dg.context import DgContext
from dagster_dg.utils import DgClickCommand, DgClickGroup, pushd
from dagster_dg.utils.filesystem import watch_paths


@click.group(name="check", cls=DgClickGroup)
def check_group():
    """Commands for checking the integrity of your Dagster code."""


# ########################
# ##### COMPONENT
# ########################


@check_group.command(name="yaml", cls=DgClickCommand)
@click.argument("paths", nargs=-1, type=click.Path(exists=True))
@click.option(
    "--watch", is_flag=True, help="Watch for changes to the component files and re-validate them."
)
@dg_global_options
def check_yaml_command(
    paths: Sequence[str],
    watch: bool,
    **global_options: object,
) -> None:
    """Check component.yaml files against their schemas, showing validation errors."""
    cli_config = normalize_cli_config(global_options, click.get_current_context())
    dg_context = DgContext.for_project_environment(Path.cwd(), cli_config)
    resolved_paths = [Path(path).absolute() for path in paths]

    def run_check(_: Any = None) -> bool:
        return check_yaml_fn(dg_context, resolved_paths)

    if watch:
        watched_paths = (
            resolved_paths or [dg_context.defs_path]
        ) + dg_context.component_registry_paths()
        watch_paths(watched_paths, run_check)
    else:
        if run_check(None):
            click.get_current_context().exit(0)
        else:
            click.get_current_context().exit(1)


@check_group.command(name="defs", cls=DgClickCommand)
@click.option(
    "--log-level",
    help="Set the log level for dagster services.",
    show_default=True,
    default="warning",
    type=click.Choice(["critical", "error", "warning", "info", "debug"], case_sensitive=False),
)
@click.option(
    "--log-format",
    type=click.Choice(["colored", "json", "rich"], case_sensitive=False),
    show_default=True,
    required=False,
    default="colored",
    help="Format of the logs for dagster services",
)
@click.option(
    "--verbose",
    "-v",
    flag_value=True,
    default=False,
    help="Show verbose error messages, including system frames in stack traces.",
)
@click.option(
    "--check-yaml/--no-check-yaml",
    flag_value=True,
    default=True,
    help="Whether to schema-check component.yaml files for the project before loading and checking all definitions.",
)
@dg_global_options
@click.pass_context
def check_definitions_command(
    context: click.Context,
    log_level: str,
    log_format: str,
    verbose: bool,
    **global_options: Mapping[str, object],
) -> None:
    """Loads and validates your Dagster definitions using a Dagster instance.

    If run inside a deployment directory, this command will launch all code locations in the
    deployment. If launched inside a code location directory, it will launch only that code
    location.

    When running, this command sets the environment variable `DAGSTER_IS_DEFS_VALIDATION_CLI=1`.
    This environment variable can be used to control the behavior of your code in validation mode.

    This command returns an exit code 1 when errors are found, otherwise an exit code 0.

    """
    cli_config = normalize_cli_config(global_options, context)
    dg_context = DgContext.for_workspace_or_project_environment(Path.cwd(), cli_config)

    forward_options = [
        *format_forwarded_option("--log-level", log_level),
        *format_forwarded_option("--log-format", log_format),
        *(["--verbose"] if verbose else []),
    ]

    run_cmds = (
        ["uv", "run", "dagster", "definitions", "validate"]
        if dg_context.is_project
        else ["uv", "tool", "run", "dagster", "definitions", "validate"]
    )

    with (
        pushd(dg_context.root_path),
        create_dagster_cli_cmd(dg_context, forward_options, run_cmds) as (
            cmd_location,
            cmd,
            workspace_file,
        ),
    ):
        if check_yaml_fn:
            overall_check_result = True
            project_dirs = (
                [project.path for project in dg_context.project_specs]
                if dg_context.is_workspace
                else [dg_context.root_path]
            )
            for project_dir in project_dirs:
                check_result = check_yaml_fn(
                    dg_context.for_project_environment(project_dir, cli_config),
                    [],
                )
                overall_check_result = overall_check_result and check_result
            if not overall_check_result:
                click.get_current_context().exit(1)
        print(f"Using {cmd_location}")  # noqa: T201
        if workspace_file:  # only non-None deployment context
            cmd.extend(["--workspace", workspace_file])

        print(" ".join(cmd))  # noqa: T201

        result = subprocess.run(cmd, check=False)
        if result.returncode != 0:
            sys.exit(result.returncode)

    click.echo("All definitions loaded successfully.")



import os
import textwrap
from collections.abc import Sequence
from typing import Any, Callable, Optional, TypeVar, Union

import click

from dagster_dg.config import DgCliConfig
from dagster_dg.utils import not_none, set_option_help_output_group

T_Command = TypeVar("T_Command", bound=Union[Callable[..., Any], click.Command])

# ########################
# ##### HELPERS
# ########################


def make_option_group(
    options_dict: dict[str, click.Option],
) -> Callable[..., Union[T_Command, Callable[[T_Command], T_Command]]]:
    def option_group(
        fn: Optional[T_Command] = None, *, names: Optional[Sequence[str]] = None
    ) -> Union[T_Command, Callable[[T_Command], T_Command]]:
        if fn:
            options = [options_dict[name] for name in names or list(options_dict.keys())]
            if isinstance(fn, click.Command):
                for option in options:
                    fn.params.append(option)
            else:
                # This is borrowed from click itself, it is how its decorators operate on both commands
                # and regular functions.
                if not hasattr(fn, "__click_params__"):
                    fn.__click_params__ = []  # type: ignore
                for option in options:
                    fn.__click_params__.append(option)  # type: ignore

            return fn
        else:
            return lambda fn: option_group(fn, names=names)  # type: ignore

    return option_group


# ########################
# ##### GLOBAL
# ########################

# Defaults are defined on the DgConfig object.
GLOBAL_OPTIONS = {
    not_none(option.name): option
    for option in [
        click.Option(
            ["--cache-dir"],
            default=str(DgCliConfig.cache_dir),
            help="Specify a directory to use for the cache.",
        ),
        click.Option(
            ["--disable-cache"],
            is_flag=True,
            default=DgCliConfig.disable_cache,
            help="Disable the cache..",
        ),
        click.Option(
            ["--verbose"],
            is_flag=True,
            default=DgCliConfig.verbose,
            help="Enable verbose output for debugging.",
        ),
        click.Option(
            ["--use-component-module", "use_component_modules"],
            type=str,
            multiple=True,
            default=DgCliConfig.__dataclass_fields__["use_component_modules"].default_factory(),
            hidden=True,
            help=textwrap.dedent("""
                Specify a list of remote environment modules expected to contain components.
                When retrieving the default set of components from the target environment, only
                components from these modules will be fetched. This overrides the default behavior
                of fetching all components registered under entry points in the remote environment.
                This is useful primarily for testing, as it allows targeting of a stable set of test
                components.
            """).strip(),
        ),
        click.Option(
            ["--use-dg-managed-environment/--no-use-dg-managed-environment"],
            is_flag=True,
            default=DgCliConfig.use_dg_managed_environment,
            help="Enable management of the virtual environment with uv.",
        ),
        click.Option(
            ["--require-local-venv/--no-require-local-venv"],
            is_flag=True,
            default=DgCliConfig.require_local_venv,
            help="Require use of a local virtual environment (`.venv` found in ancestors of the working directory).",
        ),
    ]
}

# Ensure that these options show up in the help output under the "Global options" group.
for option in GLOBAL_OPTIONS.values():
    set_option_help_output_group(option, "Global options")

dg_global_options = make_option_group(GLOBAL_OPTIONS)

# ########################
# ##### EDITABLE DAGSTER
# ########################

# When set, this will cause project scaffolding to default to --use-editable-dagster mode.
# This is a private feature designed to prevent mistakes during development.
DEFAULT_EDITABLE_DAGSTER_PROJECTS_ENV_VAR = "DG_USE_EDITABLE_DAGSTER"


# Returns false if the environment variable is not set or is set to "false".
def is_use_editable_env_var_true() -> bool:
    env_var_value = os.getenv(DEFAULT_EDITABLE_DAGSTER_PROJECTS_ENV_VAR)
    if not env_var_value:
        return False

    return env_var_value != "false"


EDITABLE_DAGSTER_OPTIONS = {
    not_none(option.name): option
    for option in [
        click.Option(
            ["--use-editable-dagster"],
            type=str,
            flag_value="TRUE",
            is_flag=False,
            default="TRUE" if is_use_editable_env_var_true() else None,
            help=(
                "Install all Dagster package dependencies from a local Dagster clone. Accepts a path to local Dagster clone root or"
                " may be set as a flag (no value is passed). If set as a flag,"
                " the location of the local Dagster clone will be read from the `DAGSTER_GIT_REPO_DIR` environment variable."
            ),
        ),
    ]
}

dg_editable_dagster_options = make_option_group(EDITABLE_DAGSTER_OPTIONS)



class DgError(Exception):
    """Base class for errors thrown by the dg CLI."""

    pass


class DgValidationError(DgError):
    """Error raised when a configuration file fails validation."""

    pass



from pathlib import Path
from typing import Final, Optional

import click

from dagster_dg.cli.scaffold import DEFAULT_WORKSPACE_NAME
from dagster_dg.cli.shared_options import dg_editable_dagster_options, dg_global_options
from dagster_dg.config import (
    DgRawWorkspaceConfig,
    DgWorkspaceScaffoldProjectOptions,
    normalize_cli_config,
)
from dagster_dg.context import DgContext
from dagster_dg.scaffold import scaffold_project, scaffold_workspace
from dagster_dg.utils import DgClickCommand, exit_with_error

# Workspace
_DEFAULT_INIT_PROJECTS_DIR: Final = "projects"


@click.command(name="init", cls=DgClickCommand)
@dg_editable_dagster_options
@dg_global_options
def init_command(
    use_editable_dagster: Optional[str],
    **global_options: object,
):
    """Initialize a new Dagster workspace and a first project within that workspace.

    The scaffolded workspace folder has the following structure:

    \b
    ├── dagster-workspace
    │   ├── projects
    |   |   └── <Dagster projects go here>
    |   ├── libraries
    |   |   └── <Shared packages go here>
    │   └── pyproject.toml

    """  # noqa: D301
    cli_config = normalize_cli_config(global_options, click.get_current_context())

    workspace_name = click.prompt(
        "Enter the name of your Dagster workspace",
        type=str,
        default=DEFAULT_WORKSPACE_NAME,
    ).strip()

    workspace_config = DgRawWorkspaceConfig(
        scaffold_project_options=DgWorkspaceScaffoldProjectOptions.get_raw_from_cli(
            use_editable_dagster,
        )
    )

    workspace_path = scaffold_workspace(workspace_name, workspace_config)
    workspace_dg_context = DgContext.from_file_discovery_and_command_line_config(
        workspace_path, cli_config
    )

    project_name = click.prompt(
        "Enter the name of your first Dagster project (or press Enter to continue without creating a project)",
        type=str,
        default="",
        show_default=False,
    ).strip()

    if not project_name:
        click.echo(
            "Continuing without adding a project. You can create one later by running `dg scaffold project`."
        )
    else:
        project_path = Path(workspace_path, _DEFAULT_INIT_PROJECTS_DIR, project_name)
        if project_path.exists():
            exit_with_error(f"A file or directory already exists at {project_path}.")
        workspace_dg_context = DgContext.from_file_discovery_and_command_line_config(
            workspace_path, cli_config
        )

        scaffold_project(
            project_path,
            workspace_dg_context,
            use_editable_dagster=use_editable_dagster,
            skip_venv=False,
            populate_cache=True,
        )

        click.echo("You can create additional projects later by running `dg scaffold project`.")



import shutil
from pathlib import Path
from typing import Final, Literal, Optional

from typing_extensions import Self, TypeAlias

from dagster_dg.config import DgConfig
from dagster_dg.utils import is_macos, is_windows

_CACHE_CONTAINER_DIR_NAME: Final = "dg-cache"

CachableDataType: TypeAlias = Literal["component_registry_data", "all_components_schema"]


def get_default_cache_dir() -> Path:
    if is_windows():
        return Path.home() / "AppData" / "dg" / "cache"
    elif is_macos():
        return Path.home() / "Library" / "Caches" / "dg"
    else:
        return Path.home() / ".cache" / "dg"


class DgCache:
    @classmethod
    def from_default(cls) -> Self:
        return cls.from_parent_path(get_default_cache_dir())

    @classmethod
    def from_config(cls, config: DgConfig) -> Self:
        return cls.from_parent_path(
            parent_path=config.cli.cache_dir,
            logging_enabled=config.cli.verbose,
        )

    # This is the preferred constructor to use when creating a cache. It ensures that all data is
    # stored inside an additional container directory inside the user-specified cache directory.
    # When we clear the cache, we only delete this container directory. This is to avoid accidents
    # when the user mistakenly specifies a cache directory that contains other data.
    @classmethod
    def from_parent_path(cls, parent_path: Path, logging_enabled: bool = False) -> Self:
        root_path = parent_path / _CACHE_CONTAINER_DIR_NAME
        return cls(root_path, logging_enabled)

    def __init__(self, root_path: Path, logging_enabled: bool):
        self._root_path = root_path
        self._root_path.mkdir(parents=True, exist_ok=True)
        self._logging_enabled = logging_enabled

    def clear_key(self, key: tuple[str, ...]) -> None:
        path = self._get_path(key)
        if path.exists():
            path.unlink()
            self.log(f"CACHE [clear-key]: {path}")

    def clear_all(self) -> None:
        shutil.rmtree(self._root_path)
        self.log(f"CACHE [clear-all]: {self._root_path}")

    def get(self, key: tuple[str, ...]) -> Optional[str]:
        path = self._get_path(key)
        if path.exists():
            self.log(f"CACHE [hit]: {path}")
            return path.read_text()
        else:
            self.log(f"CACHE [miss]: {path}")
            return None

    def set(self, key: tuple[str, ...], value: str) -> None:
        path = self._get_path(key)
        path.parent.mkdir(parents=True, exist_ok=True)
        path.write_text(value)
        self.log(f"CACHE [write]: {path}")

    def _get_path(self, key: tuple[str, ...]) -> Path:
        return Path(self._root_path, *key)

    def log(self, message: str) -> None:
        if self._logging_enabled:
            print(message)  # noqa: T201



import textwrap
from dataclasses import dataclass


def _generate_invalid_component_typename_error_message(typename: str) -> str:
    return textwrap.dedent(f"""
        Invalid component type name: `{typename}`.
        Type names must be a "."-separated string of valid Python identifiers with at least two segments.
    """)


@dataclass(frozen=True)
class ComponentKey:
    namespace: str
    name: str

    def to_typename(self) -> str:
        return f"{self.namespace}.{self.name}"

    @staticmethod
    def from_typename(typename: str) -> "ComponentKey":
        parts = typename.split(".")
        for part in parts:
            if not part.isidentifier():
                raise ValueError(_generate_invalid_component_typename_error_message(typename))
        if len(parts) < 2:
            raise ValueError(_generate_invalid_component_typename_error_message(typename))
        namespace, _, name = typename.rpartition(".")
        return ComponentKey(name=name, namespace=namespace)



from collections.abc import Mapping, Sequence
from pathlib import Path
from typing import Any, NamedTuple, Optional, Union

KeyPathSegment = Union[str, int]
KeyPath = Sequence[KeyPathSegment]


class LineCol(NamedTuple):
    line: int
    col: int


class SourcePosition(NamedTuple):
    filename: str
    start: LineCol
    end: LineCol

    def __str__(self):
        return f"{self.filename}:{self.start.line}"


class SourcePositionTree(NamedTuple):
    """Represents a tree where every node has a SourcePosition."""

    position: SourcePosition
    children: Mapping[KeyPathSegment, "SourcePositionTree"]

    def __hash__(self) -> int:
        return hash((self.position, tuple(sorted(self.children.items()))))

    def lookup(self, key_path: KeyPath) -> Optional[SourcePosition]:
        """Returns the source position of the descendant at the given path."""
        if len(key_path) == 0:
            return self.position
        head, *tail = key_path
        if head not in self.children:
            return None
        return self.children[head].lookup(tail)

    def lookup_closest_and_path(
        self, key_path: KeyPath, trace: Optional[Sequence[SourcePosition]]
    ) -> tuple[SourcePosition, Sequence[SourcePosition]]:
        """Returns the source position of the descendant at the given path. If the path does not
        exist, returns the source position of the nearest ancestor.
        """
        trace = [*trace, self.position] if trace else [self.position]
        if len(key_path) == 0:
            return self.position, trace
        head, *tail = key_path
        if head not in self.children:
            return self.position, trace
        return self.children[head].lookup_closest_and_path(tail, trace)


class ValueAndSourcePositionTree(NamedTuple):
    """A tree-like object (like a JSON-structured dict) and an accompanying SourcePositionTree.

    Each tree node in the SourcePositionTree is expected to correspond to in value.
    """

    value: Any
    source_position_tree: SourcePositionTree


class SourcePositionAndKeyPath(NamedTuple):
    """Represents a source position and key path within a file.

    Args:
        key_path (KeyPath): The path of keys that lead to the current object, where each element in
            the path is either a string (for dict keys) or an integer (for list indices).
        source_position (Optional[SourcePosition]): The source position of the object in the
            document, if available.
    """

    key_path: KeyPath
    source_position: Optional[SourcePosition]


class HasSourcePositionAndKeyPath:
    _source_position_and_key_path: Optional[SourcePositionAndKeyPath] = None

    @property
    def source_position(self) -> SourcePosition:
        """Returns the underlying source position of the object, including
        the source file and line number.
        """
        assert self._source_position_and_key_path
        assert self._source_position_and_key_path.source_position
        return self._source_position_and_key_path.source_position

    @property
    def source_file(self) -> Path:
        """Path to the source file where the object is defined."""
        assert self._source_position_and_key_path
        return Path(self.source_position.filename)

    @property
    def source_file_name(self) -> str:
        """Name of the source file where the object is defined."""
        return self.source_file.name



import subprocess
from collections.abc import Mapping
from pathlib import Path
from typing import Optional

import click

from dagster_dg.cli.dev import format_forwarded_option
from dagster_dg.cli.shared_options import dg_global_options
from dagster_dg.config import normalize_cli_config
from dagster_dg.context import DgContext
from dagster_dg.utils import DgClickCommand


@click.command(name="launch", cls=DgClickCommand)
@click.option("--assets", help="Comma-separated Asset selection to target", required=True)
@click.option("--partition", help="Asset partition to target", required=False)
@click.option(
    "--partition-range",
    help="Asset partition range to target i.e. <start>...<end>",
    required=False,
)
@click.option(
    "--config-json", type=click.STRING, help="JSON string of config to use for the launched run."
)
@dg_global_options
def launch_command(
    assets: str,
    partition: Optional[str],
    partition_range: Optional[str],
    config_json: Optional[str],
    **global_options: Mapping[str, object],
):
    """Launch a Dagster run."""
    forward_options = [
        *format_forwarded_option("--select", assets),
        *format_forwarded_option("--partition", partition),
        *format_forwarded_option("--partition-range", partition_range),
        *format_forwarded_option("--config-json", config_json),
    ]

    cli_config = normalize_cli_config(global_options, click.get_current_context())

    # TODO - make this work in a workspace and/or cloud context instead of materializing the
    # assets in process. It should use the instance's run launcher to launch a run (or backfill
    # depending on the asset selection) and optionally stream logs from that run or backfill in
    # the command line until it finishes. right now, it only works in a project context in local
    # dev/OSS, and executes the selected assets in process using the "dagster asset materialize"
    # command.

    dg_context = DgContext.for_project_environment(Path.cwd(), cli_config)

    cmd_location = dg_context.get_executable("dagster")
    click.echo(f"Using {cmd_location}")

    args = [
        "--working-directory",
        str(dg_context.root_path),
        "--module-name",
        str(dg_context.code_location_target_module_name),
    ]

    result = subprocess.run(
        [cmd_location, "asset", "materialize", *args, *forward_options], check=False
    )
    if result.returncode != 0:
        click.echo("Failed to launch assets.")
        click.get_current_context().exit(result.returncode)



import re
from collections.abc import Sequence
from typing import Optional, Union

import click
import typer
from jsonschema import ValidationError

from dagster_dg.component_key import ComponentKey
from dagster_dg.yaml_utils.source_position import SourcePositionTree


@click.group(name="check")
def check_cli():
    """Commands for checking components."""


def prepend_lines_with_line_numbers(
    lines_with_numbers: Sequence[tuple[Optional[int], str]],
) -> Sequence[str]:
    """Prepend each line with a line number, right-justified to the maximum line number length.

    Args:
        lines_with_numbers: A sequence of tuples, where the first element is the line number and the
            second element is the line content. Some lines may have a `None` line number, which
            will be rendered as an empty string, used for e.g. inserted error message lines.
    """
    max_line_number_length = max([len(str(n)) for n, _ in lines_with_numbers])
    return [
        f"{(str(n) if n else '').rjust(max_line_number_length)} | {line.rstrip()}"
        for n, line in lines_with_numbers
    ]


def augment_inline_error_message(location: str, msg: str):
    """Improves a subset of Pyright error messages by including location information."""
    last_location_part = location.split(".")[-1]
    if msg == "Field required":
        return f"Field `{last_location_part}` is required but not provided"
    return msg


ADDITIONAL_PROPERTIES_ERROR_MESSAGE = (
    r"Additional properties are not allowed \('(.*?)' was unexpected\)"
)


def augment_error_path(error_details: ValidationError) -> Sequence[Union[str, int]]:
    """Augment the error location (e.g. key) for certain error messages.

    In particular, for extra properties, returns the location of the extra property instead
    of the parent key.
    """
    additional_property = re.match(ADDITIONAL_PROPERTIES_ERROR_MESSAGE, error_details.message)
    if additional_property:
        return [*error_details.absolute_path, additional_property.group(1)]
    return error_details.absolute_path


def format_indented_error_msg(col: int, msg: str) -> str:
    """Format an error message with a caret pointing to the column where the error occurred."""
    return typer.style(" " * (col - 1) + f"^ {msg}", fg=typer.colors.YELLOW)


OFFSET_LINES_BEFORE = 2
OFFSET_LINES_AFTER = 3


def error_dict_to_formatted_error(
    component_key: Optional[ComponentKey],
    error_details: ValidationError,
    source_position_tree: SourcePositionTree,
    prefix: Sequence[str] = (),
) -> str:
    """Convert a ValidationError to a formatted error message, including
    a code snippet of the offending YAML file.

    Args:
        component_name: The name of the component that the error occurred in, e.g. "my_component".
        error_details: The JSON Schema ValidationError object.
        source_position_tree: The SourcePositionTree object, which contains the source position of
            each line in the YAML file.
        prefix: A prefix to the JSON path of the location of the error in the YAML file. Used because
            we validate attributes separately from the top-level component YAML fields, so this is often
            set to e.g. ["attributes"] when validating the internal attributes of a component.
    """
    error_path = augment_error_path(error_details)

    source_position, source_position_path = source_position_tree.lookup_closest_and_path(
        [*prefix, *error_path], trace=None
    )

    # Retrieves dotted path representation of the location of the error in the YAML file, e.g.
    # attributes.nested.foo.an_int
    location = ".".join([*prefix, *[str(part) for part in error_path]]).split(" at ")[0]

    # Find the first source position that has a different start line than the current source position
    # This is e.g. the parent json key of the current source position
    preceding_source_position = next(
        iter(
            [
                value
                for value in reversed(list(source_position_path))
                if value.start.line < source_position.start.line
            ]
        ),
        source_position,
    )
    with open(source_position.filename) as f:
        lines = f.readlines()
        lines_with_line_numbers = list(zip(range(1, len(lines) + 1), lines))

        filtered_lines_with_line_numbers = (
            lines_with_line_numbers[
                max(
                    0, preceding_source_position.start.line - OFFSET_LINES_BEFORE
                ) : source_position.start.line
            ]
            + [
                (
                    None,
                    format_indented_error_msg(
                        source_position.start.col,
                        augment_inline_error_message(location, error_details.message),
                    ),
                )
            ]
            + lines_with_line_numbers[
                source_position.start.line : source_position.end.line + OFFSET_LINES_AFTER
            ]
        )
        # Combine the filtered lines with the line numbers, and add empty lines before and after
        lines_with_line_numbers = prepend_lines_with_line_numbers(
            [(None, ""), *filtered_lines_with_line_numbers, (None, "")]
        )
        code_snippet = "\n".join(lines_with_line_numbers)

    fmt_filename = (
        f"{source_position.filename}"
        f":{typer.style(source_position.start.line, fg=typer.colors.GREEN)}"
    )
    fmt_location = typer.style(location, fg=typer.colors.BRIGHT_WHITE)
    fmt_name = typer.style(
        f"{component_key.to_typename()} " if component_key else "", fg=typer.colors.RED
    )
    return f"{fmt_filename} - {fmt_name}{fmt_location} {error_details.message}\n{code_snippet}\n"



import datetime
import hashlib
import time
from collections.abc import Sequence
from pathlib import Path
from typing import Any, Callable, Optional

from watchdog.events import FileSystemEvent, FileSystemEventHandler
from watchdog.observers import Observer

from dagster_dg.utils import (
    DEFAULT_FILE_EXCLUDE_PATTERNS,
    clear_screen,
    hash_directory_metadata,
    hash_file_metadata,
)


def hash_paths(
    paths: Sequence[Path],
    includes: Optional[Sequence[str]] = None,
    excludes: Sequence[str] = DEFAULT_FILE_EXCLUDE_PATTERNS,
    error_on_missing: bool = True,
) -> str:
    """Hash the given paths, including their metadata.

    Args:
        paths: The paths to hash.
        includes: A list of glob patterns to target, excluding files that don't match any of the patterns.
        excludes: A list of glob patterns to exclude, including files that match any of the patterns. Defaults to
            various Python-related files that are not relevant to the contents of the project.

    Returns:
        The hash of the paths.
    """
    hasher = hashlib.md5()
    for path in paths:
        if path.is_dir():
            hash_directory_metadata(
                hasher,
                path,
                includes=includes,
                excludes=excludes,
                error_on_missing=error_on_missing,
            )
        else:
            hash_file_metadata(hasher, path, error_on_missing=error_on_missing)
    return hasher.hexdigest()


class PathChangeHandler(FileSystemEventHandler):
    """Basic handler that clears the screen and re-executes the callback when any of the watched paths change
    in a way that would affect the hash of the paths. Passes the new hash to the callback.
    """

    def __init__(
        self,
        paths: Sequence[Path],
        includes: Optional[Sequence[str]],
        excludes: Sequence[str],
        callback: Callable[[str], Any],
    ):
        self._callback = callback
        self._paths = paths
        self._includes = includes
        self._excludes = excludes
        self._prev_hash = hash_paths(self._paths, self._includes, self._excludes)
        self.clear_and_execute(self._prev_hash)

    def dispatch(self, _event: FileSystemEvent):
        new_hash = hash_paths(self._paths, self._includes, self._excludes, error_on_missing=False)

        if new_hash != self._prev_hash:
            self.clear_and_execute(new_hash)
            self._prev_hash = new_hash

    def clear_and_execute(self, new_hash: str):
        clear_screen()
        self._callback(new_hash)
        current_time = datetime.datetime.now().strftime("%H:%M:%S")
        print(f"\nUpdated at {current_time}, watching for changes...")  # noqa: T201


# This is a global variable that is used to signal the watcher to exit in tests
SHOULD_WATCHER_EXIT = False


def watch_paths(
    paths: Sequence[Path],
    callback: Callable[[str], Any],
    includes: Optional[Sequence[str]] = None,
    excludes: Sequence[str] = DEFAULT_FILE_EXCLUDE_PATTERNS,
):
    """Watches the given paths for changes and calls the callback when they change.
    The callback should take a single argument, the new hash of the paths.

    Runs synchronously until the observer is stopped, or keyboard interrupt is received.

    Args:
        paths: The paths to watch.
        callback: The callback to call when path contents change.
        includes: A list of glob patterns to target, excluding files that don't match any of the patterns.
        excludes: A list of glob patterns to exclude, including files that match any of the patterns. Defaults to
            various Python-related files that are not relevant to the contents of the project.
    """
    observer = Observer()
    handler = PathChangeHandler(paths, includes, excludes, callback)
    for path in paths:
        observer.schedule(handler, str(path), recursive=True)
    observer.start()
    try:
        while observer.is_alive() and not SHOULD_WATCHER_EXIT:
            time.sleep(0.5)
    finally:
        observer.stop()
        observer.join()



import functools
from collections.abc import Mapping, Sequence
from dataclasses import dataclass, field
from pathlib import Path
from typing import (
    Any,
    Callable,
    Literal,
    Optional,
    TypedDict,
    TypeVar,
    Union,
    cast,
    get_args,
    get_origin,
)

import click
import tomlkit
import tomlkit.items
from click.core import ParameterSource
from typing_extensions import Never, NotRequired, Required, Self, TypeAlias, TypeGuard

from dagster_dg.error import DgError, DgValidationError
from dagster_dg.utils import get_toml_node, is_macos, is_windows

T = TypeVar("T")

DEFAULT_BUILTIN_COMPONENT_LIB = "dagster_components"


def _get_default_cache_dir() -> Path:
    if is_windows():
        return Path.home() / "AppData" / "dg" / "cache"
    elif is_macos():
        return Path.home() / "Library" / "Caches" / "dg"
    else:
        return Path.home() / ".cache" / "dg"


DEFAULT_CACHE_DIR = _get_default_cache_dir()


def discover_workspace_root(path: Path) -> Optional[Path]:
    workspace_config_path = discover_config_file(
        path, lambda config: config["directory_type"] == "workspace"
    )
    return workspace_config_path.parent if workspace_config_path else None


def discover_config_file(
    path: Path,
    predicate: Optional[Callable[[Mapping[str, Any]], bool]] = None,
) -> Optional[Path]:
    current_path = path.absolute()
    while True:
        config_path = current_path / "pyproject.toml"
        if config_path.exists() and has_dg_file_config(config_path, predicate):
            return config_path
        if current_path == current_path.parent:  # root
            return
        current_path = current_path.parent


# NOTE: The set/has/get_config_from_cli_context is only used for the dynamically generated scaffold
# component commands. Hopefully we can get rid of it in the future.

_CLI_CONTEXT_CONFIG_KEY = "config"


def set_config_on_cli_context(cli_context: click.Context, config: "DgRawCliConfig") -> None:
    cli_context.ensure_object(dict)
    cli_context.obj[_CLI_CONTEXT_CONFIG_KEY] = config


def has_config_on_cli_context(cli_context: click.Context) -> bool:
    return _CLI_CONTEXT_CONFIG_KEY in cli_context.ensure_object(dict)


def get_config_from_cli_context(cli_context: click.Context) -> "DgRawCliConfig":
    cli_context.ensure_object(dict)
    return cli_context.obj[_CLI_CONTEXT_CONFIG_KEY]


# ########################
# ##### MAIN
# ########################


@dataclass
class DgConfig:
    cli: "DgCliConfig"
    project: Optional["DgProjectConfig"] = None
    workspace: Optional["DgWorkspaceConfig"] = None

    @classmethod
    def default(cls) -> Self:
        return cls(DgCliConfig.default())

    # Note that this function takes an argument called `container_workspace_file_config` instead of
    # just `workspace_file_config` because this is only to be passed when the root file config
    # points to a project inside of a workspace. If there is no package and the
    # root is itself a workspace, then `root_file_config` corresponds to the workspace and
    # `container_workspace_file_config` should be None.
    @classmethod
    def from_partial_configs(
        cls,
        root_file_config: Optional["DgFileConfig"] = None,
        container_workspace_file_config: Optional["DgWorkspaceFileConfig"] = None,
        command_line_config: Optional["DgRawCliConfig"] = None,
    ) -> Self:
        cli_partials: list[DgRawCliConfig] = []

        if container_workspace_file_config:
            if "cli" in container_workspace_file_config:
                cli_partials.append(container_workspace_file_config["cli"])
            workspace_config = DgWorkspaceConfig.from_raw(
                container_workspace_file_config["workspace"]
            )

        if root_file_config:
            if "cli" in root_file_config:
                cli_partials.append(root_file_config["cli"])

            if is_workspace_file_config(root_file_config):
                project_config = None
                workspace_config = DgWorkspaceConfig.from_raw(root_file_config["workspace"])
            elif is_project_file_config(root_file_config):
                project_config = DgProjectConfig.from_raw(root_file_config["project"])
                if not container_workspace_file_config:
                    workspace_config = None

        else:
            project_config = None
            workspace_config = None

        if command_line_config:
            cli_partials.append(command_line_config)
        cli_config = DgCliConfig.from_raw(*cli_partials) if cli_partials else DgCliConfig.default()

        return cls(cli_config, project_config, workspace_config)


# ########################
# ##### CLI
# ########################


@dataclass
class DgCliConfig:
    """CLI configuration for Dg.

    Args:
        disable_cache (bool): If True, disable caching. Defaults to False.
        cache_dir (Optional[str]): The directory to use for caching. If None, the default cache will
            be used.
        verbose (bool): If True, log debug information.
        use_component_modules (list[str]): Specify a list of modules containing components.
            Any components retrieved from the remote environment will be filtered to only include those
            from these modules. This is useful primarily for testing, as it allows targeting of a stable
            set of test components.
        use_dg_managed_environment (bool): If True, `dg` will build and manage a virtual environment
            using `uv`. Note that disabling the managed enviroment will also disable caching.
        require_local_venv (bool): If True, commands that access an environment with
            dagster-components will only use a `.venv` directory discovered in the ancestor tree. If no
            `.venv` directory is discovered, an error will be raised. Note that this disallows the use
            of both the system python environment and non-local but activated virtual environments.
    """

    disable_cache: bool = False
    cache_dir: Path = DEFAULT_CACHE_DIR
    verbose: bool = False
    use_component_modules: list[str] = field(default_factory=list)
    use_dg_managed_environment: bool = True
    require_local_venv: bool = True

    @classmethod
    def default(cls) -> Self:
        return cls()

    @classmethod
    def from_raw(cls, *partials: "DgRawCliConfig") -> Self:
        merged = cast("DgRawCliConfig", functools.reduce(lambda acc, x: {**acc, **x}, partials))  # type: ignore
        return cls(
            disable_cache=merged.get("disable_cache", DgCliConfig.disable_cache),
            cache_dir=Path(merged["cache_dir"]) if "cache_dir" in merged else DgCliConfig.cache_dir,
            verbose=merged.get("verbose", DgCliConfig.verbose),
            use_component_modules=merged.get(
                "use_component_modules",
                cls.__dataclass_fields__["use_component_modules"].default_factory(),
            ),
            use_dg_managed_environment=merged.get(
                "use_dg_managed_environment", DgCliConfig.use_dg_managed_environment
            ),
            require_local_venv=merged.get("require_local_venv", DgCliConfig.require_local_venv),
        )


# All fields are optional
class DgRawCliConfig(TypedDict, total=False):
    disable_cache: bool
    cache_dir: str
    verbose: bool
    use_component_modules: Sequence[str]
    use_dg_managed_environment: bool
    require_local_venv: bool


# ########################
# ##### PROJECT
# ########################


@dataclass
class DgProjectConfig:
    root_module: str
    defs_module: Optional[str] = None
    code_location_target_module: Optional[str] = None
    code_location_name: Optional[str] = None

    @classmethod
    def from_raw(cls, raw: "DgRawProjectConfig") -> Self:
        return cls(
            root_module=raw["root_module"],
            defs_module=raw.get("defs_module", DgProjectConfig.defs_module),
            code_location_name=raw.get("code_location_name", DgProjectConfig.code_location_name),
            code_location_target_module=raw.get(
                "code_location_target_module",
                DgProjectConfig.code_location_target_module,
            ),
        )


class DgRawProjectConfig(TypedDict):
    root_module: Required[str]
    defs_module: NotRequired[str]
    code_location_target_module: NotRequired[str]
    code_location_name: NotRequired[str]


# ########################
# ##### WORKSPACE
# ########################


@dataclass
class DgWorkspaceConfig:
    projects: list["DgWorkspaceProjectSpec"]
    scaffold_project_options: "DgWorkspaceScaffoldProjectOptions"

    @classmethod
    def from_raw(cls, raw: "DgRawWorkspaceConfig") -> Self:
        projects = [DgWorkspaceProjectSpec.from_raw(spec) for spec in raw.get("projects", [])]
        scaffold_project_options = DgWorkspaceScaffoldProjectOptions.from_raw(
            raw.get("scaffold_project_options", {})
        )
        return cls(projects, scaffold_project_options)


class DgRawWorkspaceConfig(TypedDict, total=False):
    projects: list["DgRawWorkspaceProjectSpec"]
    scaffold_project_options: "DgRawWorkspaceNewProjectOptions"


@dataclass
class DgWorkspaceProjectSpec:
    path: Path
    code_location_name: Optional[str] = None

    @classmethod
    def from_raw(cls, raw: "DgRawWorkspaceProjectSpec") -> Self:
        return cls(
            path=Path(raw["path"]),
            code_location_name=raw.get(
                "code_location_name", DgWorkspaceProjectSpec.code_location_name
            ),
        )


class DgRawWorkspaceProjectSpec(TypedDict, total=False):
    path: Required[str]
    code_location_name: str


@dataclass
class DgWorkspaceScaffoldProjectOptions:
    use_editable_dagster: Union[str, bool] = False

    @classmethod
    def from_raw(cls, raw: "DgRawWorkspaceNewProjectOptions") -> Self:
        return cls(
            use_editable_dagster=raw.get(
                "use_editable_dagster", DgWorkspaceScaffoldProjectOptions.use_editable_dagster
            ),
        )

    # This is here instead of on `DgRawWorkspaceNewProjectOptions` because TypedDict can't have
    # methods.
    @classmethod
    def get_raw_from_cli(
        cls,
        use_editable_dagster: Optional[str],
    ) -> "DgRawWorkspaceNewProjectOptions":
        raw_scaffold_project_options: DgRawWorkspaceNewProjectOptions = {}
        if use_editable_dagster:
            raw_scaffold_project_options["use_editable_dagster"] = (
                True if use_editable_dagster == "TRUE" else use_editable_dagster
            )
        return raw_scaffold_project_options


class DgRawWorkspaceNewProjectOptions(TypedDict, total=False):
    use_editable_dagster: Union[str, bool]


# ########################
# ##### CLI CONFIG
# ########################


def normalize_cli_config(
    cli_options: Mapping[str, object], cli_context: click.Context
) -> DgRawCliConfig:
    # Remove any options that weren't explicitly provided.
    filtered_options = {
        key: value
        for key, value in cli_options.items()
        if cli_context.get_parameter_source(key) != ParameterSource.DEFAULT
    }
    return _validate_cli_config(filtered_options)


def _validate_cli_config(cli_opts: Mapping[str, object]) -> DgRawCliConfig:
    try:
        for key, type_ in DgRawCliConfig.__annotations__.items():
            _validate_cli_config_setting(cli_opts, key, type_)
        _validate_cli_config_no_extraneous_keys(cli_opts)
    except DgValidationError as e:
        _raise_cli_config_validation_error(str(e))
    return cast(DgRawCliConfig, cli_opts)


def _validate_cli_config_setting(cli_opts: Mapping[str, object], key: str, type_: type) -> None:
    if key in cli_opts and not _match_type(cli_opts[key], type_):
        raise DgValidationError(f"`{key}` must be a {type_.__name__}.")


def _validate_cli_config_no_extraneous_keys(cli_opts: Mapping[str, object]) -> None:
    extraneous_keys = [k for k in cli_opts.keys() if k not in DgRawCliConfig.__annotations__]
    if extraneous_keys:
        raise DgValidationError(f"Unrecognized fields: {extraneous_keys}")


def _raise_cli_config_validation_error(message: str) -> None:
    raise DgError(f"Error in CLI options: {message}")


# ########################
# ##### FILE CONFIG
# ########################

DgFileConfigDirectoryType = Literal["workspace", "project"]


class DgWorkspaceFileConfig(TypedDict):
    directory_type: Required[Literal["workspace"]]
    workspace: Required[DgRawWorkspaceConfig]
    cli: NotRequired[DgRawCliConfig]


def is_workspace_file_config(config: "DgFileConfig") -> TypeGuard[DgWorkspaceFileConfig]:
    return config["directory_type"] == "workspace"


class DgProjectFileConfig(TypedDict):
    directory_type: Required[Literal["project"]]
    project: Required[DgRawProjectConfig]
    cli: NotRequired[DgRawCliConfig]


def is_project_file_config(config: "DgFileConfig") -> TypeGuard[DgProjectFileConfig]:
    return config["directory_type"] == "project"


DgFileConfig: TypeAlias = Union[DgWorkspaceFileConfig, DgProjectFileConfig]


def has_dg_file_config(
    path: Path, predicate: Optional[Callable[[Mapping[str, Any]], bool]] = None
) -> bool:
    toml = tomlkit.parse(path.read_text()).unwrap()
    return "dg" in toml.get("tool", {}) and (predicate(toml["tool"]["dg"]) if predicate else True)


def load_dg_root_file_config(path: Path) -> DgFileConfig:
    return _load_dg_file_config(path)


def load_dg_workspace_file_config(path: Path) -> "DgWorkspaceFileConfig":
    config = _load_dg_file_config(path)
    if is_workspace_file_config(config):
        return config
    else:
        _raise_file_config_validation_error("Expected a workspace configuration.", path)


def _load_dg_file_config(path: Path) -> DgFileConfig:
    toml = tomlkit.parse(path.read_text())
    raw_dict = get_toml_node(toml, ("tool", "dg"), tomlkit.items.Table).unwrap()
    try:
        config = _validate_dg_file_config({k: v for k, v in raw_dict.items()})
    except DgValidationError as e:
        _raise_file_config_validation_error(str(e), path)
    return config


def _validate_dg_file_config(raw_dict: Mapping[str, object]) -> DgFileConfig:
    _validate_file_config_setting(
        raw_dict,
        "directory_type",
        Required[Literal["workspace", "project"]],
        "tool.dg",
    )
    _validate_dg_config_file_cli_section(raw_dict.get("cli", {}))
    if raw_dict["directory_type"] == "workspace":
        _validate_file_config_setting(raw_dict, "workspace", dict, "tool.dg")
        _validate_file_config_workspace_section(raw_dict.get("workspace", {}))
        _validate_file_config_no_extraneous_keys(
            set(DgWorkspaceFileConfig.__annotations__.keys()), raw_dict, "tool.dg"
        )
        return cast(DgWorkspaceFileConfig, raw_dict)
    elif raw_dict["directory_type"] == "project":
        _validate_file_config_setting(raw_dict, "project", dict, "tool.dg")
        _validate_file_config_project_section(raw_dict.get("project", {}))
        _validate_file_config_no_extraneous_keys(
            set(DgProjectFileConfig.__annotations__.keys()), raw_dict, "tool.dg"
        )
        return cast(DgProjectFileConfig, raw_dict)
    else:
        raise DgError("Unreachable")


def _validate_dg_config_file_cli_section(section: object) -> None:
    if not isinstance(section, dict):
        _raise_mistyped_key_error("tool.dg.cli", get_type_str(dict), section)
    for key, type_ in DgRawCliConfig.__annotations__.items():
        _validate_file_config_setting(section, key, type_, "tool.dg.cli")
    _validate_file_config_no_extraneous_keys(
        set(DgRawCliConfig.__annotations__.keys()), section, "tool.dg.cli"
    )


def _validate_file_config_project_section(section: object) -> None:
    if not isinstance(section, dict):
        _raise_mistyped_key_error("tool.dg.project", get_type_str(dict), section)
    for key, type_ in DgRawProjectConfig.__annotations__.items():
        _validate_file_config_setting(section, key, type_, "tool.dg.project")
    _validate_file_config_no_extraneous_keys(
        set(DgRawProjectConfig.__annotations__.keys()), section, "tool.dg.project"
    )


def _validate_file_config_workspace_section(section: object) -> None:
    if not isinstance(section, dict):
        _raise_mistyped_key_error("tool.dg.workspace", get_type_str(dict), section)
    for key, type_ in DgRawWorkspaceConfig.__annotations__.items():
        if key == "projects":
            _validate_file_config_setting(section, key, list, "tool.dg.workspace")
            for i, spec in enumerate(section.get("projects") or []):
                _validate_file_config_workspace_project_spec(spec, i)
        elif key == "scaffold_project_options":
            _validate_file_config_workspace_scaffold_project_options(
                section.get("scaffold_project_options", {})
            )
        else:
            _validate_file_config_setting(section, key, type_, "tool.dg.workspace")
    _validate_file_config_no_extraneous_keys(
        set(DgRawWorkspaceConfig.__annotations__.keys()), section, "tool.dg.workspace"
    )


def _validate_file_config_workspace_project_spec(section: object, index: int) -> None:
    if not isinstance(section, dict):
        _raise_mistyped_key_error(
            f"tool.dg.workspace.projects[{index}]", get_type_str(dict), section
        )
    for key, type_ in DgRawWorkspaceProjectSpec.__annotations__.items():
        _validate_file_config_setting(section, key, type_, f"tool.dg.workspace.projects[{index}]")
    _validate_file_config_no_extraneous_keys(
        set(DgRawWorkspaceProjectSpec.__annotations__.keys()),
        section,
        f"tool.dg.workspace.projects[{index}]",
    )


def _validate_file_config_workspace_scaffold_project_options(section: object) -> None:
    if not isinstance(section, dict):
        _raise_mistyped_key_error(
            "tool.dg.workspace.scaffold_project_options", get_type_str(dict), section
        )
    for key, type_ in DgRawWorkspaceNewProjectOptions.__annotations__.items():
        _validate_file_config_setting(
            section, key, type_, "tool.dg.workspace.scaffold_project_options"
        )
    _validate_file_config_no_extraneous_keys(
        set(DgRawWorkspaceNewProjectOptions.__annotations__.keys()),
        section,
        "tool.dg.workspace.scaffold_project_options",
    )


def _validate_file_config_no_extraneous_keys(
    valid_keys: set[str], section: Mapping[str, object], toml_path: str
) -> None:
    extraneous_keys = [k for k in section.keys() if k not in valid_keys]
    if extraneous_keys:
        raise DgValidationError(f"Unrecognized fields in `{toml_path}`: {extraneous_keys}")


# expected_type Any to handle typing constructs (`Literal` etc)
def _validate_file_config_setting(
    section: Mapping[str, object],
    key: str,
    type_: Any,
    path_prefix: Optional[str] = None,
) -> None:
    origin = get_origin(type_)
    is_required = origin is Required
    class_ = type_ if origin not in (Required, NotRequired) else get_args(type_)[0]

    error_type = None
    if is_required and key not in section:
        error_type = "required"
    if key in section and not _match_type(section[key], class_):
        error_type = "mistype"
    if error_type:
        full_key = f"{path_prefix}.{key}" if path_prefix else key
        type_str = get_type_str(class_)
        if error_type == "required":
            _raise_missing_required_key_error(full_key, type_str)
        if error_type == "mistype":
            _raise_mistyped_key_error(full_key, type_str, section[key])


def _raise_missing_required_key_error(key: str, type_str: str) -> Never:
    raise DgValidationError(f"Missing required value for `{key}`. Expected {type_str}.")


def _raise_mistyped_key_error(key: str, type_str: str, value: object) -> Never:
    raise DgValidationError(f"`Invalid value for `{key}`. Expected {type_str}, got `{value}`.")


def _raise_file_config_validation_error(message: str, file_path: Path) -> Never:
    raise DgError(f"Error in configuration file {file_path}: {message}")


# expected_type Any to handle typing constructs (`Literal` etc)
def get_type_str(t: Any) -> str:
    origin = get_origin(t)
    if origin is None:
        # It's a builtin or normal class
        if hasattr(t, "__name__"):
            return t.__name__  # e.g. 'int', 'bool'
        return str(t)  # Fallback
    else:
        # It's a parametric type like list[str], Union[int, str], etc.
        args = get_args(t)
        arg_strs = sorted([get_type_str(a) for a in args])
        if origin is Union:
            return " | ".join(arg_strs)
        if origin is Literal:
            arg_strs = [f'"{a}"' for a in arg_strs]
            return " | ".join(arg_strs)
        else:
            return f"{_get_origin_name(origin)}[{', '.join(arg_strs)}]"


def _get_origin_name(origin: Any) -> str:
    if origin is Sequence:
        return "Sequence"  # avoid the collections.abc prefix
    else:
        return origin.__name__


def _match_type(obj: object, type_: Any) -> bool:
    origin = get_origin(type_)
    # If typ is not a generic alias, do a normal isinstance check
    if origin is None:
        # Edge case: Union can appear as typing.Union without origin in older versions,
        # but with modern Python, get_origin should handle it.
        # If we get here, it's a concrete type like `int`, `str`, or type(None).
        return isinstance(obj, type_)

    # Handle Union (e.g. Union[int, str])
    if origin is Union:
        subtypes = get_args(type_)  # e.g. (int, str)
        return any(_match_type(obj, st) for st in subtypes)

    # Handle Literal (e.g. Literal[3, 5, "hello"])
    if origin is Literal:
        # get_args(typ) will be the allowed literal values
        allowed_values = get_args(type_)  # e.g. (3, 5, "hello")
        return obj in allowed_values

    # Handle list[...] (e.g. list[str])
    if origin is Sequence:
        (item_type,) = get_args(type_)  # e.g. (str,) for list[str]
        if not isinstance(obj, Sequence):
            return False
        return all(_match_type(item, item_type) for item in obj)

    # Handle list[...] (e.g. list[str])
    if origin is list:
        (item_type,) = get_args(type_)  # e.g. (str,) for list[str]
        if not isinstance(obj, list):
            return False
        return all(_match_type(item, item_type) for item in obj)

    # Handle tuple[...] (e.g. tuple[int, str], tuple[str, ...])
    if origin is tuple:
        arg_types = get_args(type_)
        if not isinstance(obj, tuple):
            return False
        # Distinguish fixed-length vs variable-length (ellipsis) tuples
        if len(arg_types) == 2 and arg_types[1] is Ellipsis:
            # e.g. tuple[str, ...]
            elem_type = arg_types[0]
            return all(_match_type(item, elem_type) for item in obj)
        else:
            # e.g. tuple[int, str, float]
            if len(obj) != len(arg_types):
                return False
            return all(_match_type(item, t) for item, t in zip(obj, arg_types))

    # Handle dict[...] (e.g. dict[str, int])
    if origin is dict:
        key_type, val_type = get_args(type_)
        if not isinstance(obj, dict):
            return False
        return all(_match_type(k, key_type) and _match_type(v, val_type) for k, v in obj.items())

    # Extend with other generic types (set, frozenset, etc.) if needed
    raise NotImplementedError(f"No handler for {type_}")



import json
from dataclasses import asdict
from pathlib import Path
from typing import Any

import click
from rich.console import Console
from rich.table import Table

from dagster_dg.cli.scaffold import SHIM_COMPONENTS
from dagster_dg.cli.shared_options import dg_global_options
from dagster_dg.component import RemoteComponentRegistry
from dagster_dg.config import normalize_cli_config
from dagster_dg.context import DgContext
from dagster_dg.defs import (
    DgAssetMetadata,
    DgDefinitionMetadata,
    DgJobMetadata,
    DgScheduleMetadata,
    DgSensorMetadata,
)
from dagster_dg.error import DgError
from dagster_dg.utils import DgClickCommand, DgClickGroup


@click.group(name="list", cls=DgClickGroup)
def list_group():
    """Commands for listing Dagster entities."""


# ########################
# ##### PROJECT
# ########################


@list_group.command(name="project", cls=DgClickCommand)
@dg_global_options
def list_project_command(**global_options: object) -> None:
    """List projects in the current workspace."""
    cli_config = normalize_cli_config(global_options, click.get_current_context())
    dg_context = DgContext.for_workspace_environment(Path.cwd(), cli_config)

    for project in dg_context.project_specs:
        click.echo(project.path)


# ########################
# ##### COMPONENT
# ########################


@list_group.command(name="component", cls=DgClickCommand)
@dg_global_options
def list_component_command(**global_options: object) -> None:
    """List Dagster component instances defined in the current project."""
    cli_config = normalize_cli_config(global_options, click.get_current_context())
    dg_context = DgContext.for_project_environment(Path.cwd(), cli_config)

    for component_instance_name in dg_context.get_component_instance_names():
        click.echo(component_instance_name)


# ########################
# ##### COMPONENT TYPE
# ########################


@list_group.command(name="component-type", cls=DgClickCommand)
@click.option(
    "--json",
    "output_json",
    is_flag=True,
    default=False,
    help="Output as JSON instead of a table.",
)
@dg_global_options
def list_component_type_command(output_json: bool, **global_options: object) -> None:
    """List registered Dagster components in the current project environment."""
    cli_config = normalize_cli_config(global_options, click.get_current_context())
    dg_context = DgContext.for_defined_registry_environment(Path.cwd(), cli_config)
    registry = RemoteComponentRegistry.from_dg_context(dg_context)

    sorted_keys = sorted(registry.keys() - SHIM_COMPONENTS.keys(), key=lambda k: k.to_typename())

    # JSON
    if output_json:
        output: list[dict[str, object]] = []
        for key in sorted_keys:
            component_type_metadata = registry.get(key)
            output.append(
                {
                    "key": key.to_typename(),
                    "summary": component_type_metadata.summary,
                }
            )
        click.echo(json.dumps(output, indent=4))

    # TABLE
    else:
        table = Table(border_style="dim")
        table.add_column("Component Type", style="bold cyan", no_wrap=True)
        table.add_column("Summary")
        for key in sorted_keys:
            table.add_row(key.to_typename(), registry.get(key).summary)
        console = Console()
        console.print(table)


# ########################
# ##### DEFS
# ########################


@list_group.command(name="defs", cls=DgClickCommand)
@click.option(
    "--json",
    "output_json",
    is_flag=True,
    default=False,
    help="Output as JSON instead of a table.",
)
@dg_global_options
def list_defs_command(output_json: bool, **global_options: object) -> None:
    """List registered Dagster definitions in the current project environment."""
    cli_config = normalize_cli_config(global_options, click.get_current_context())
    dg_context = DgContext.for_project_environment(Path.cwd(), cli_config)

    result = dg_context.external_components_command(
        ["list", "definitions", "-m", dg_context.code_location_target_module_name]
    )
    definitions = [_resolve_definition(x) for x in json.loads(result)]

    # JSON
    if output_json:  # pass it straight through
        json_output = [asdict(defn) for defn in definitions]
        click.echo(json.dumps(json_output, indent=4))

    # TABLE
    else:
        assets = [item for item in definitions if isinstance(item, DgAssetMetadata)]
        jobs = [item for item in definitions if isinstance(item, DgJobMetadata)]
        schedules = [item for item in definitions if isinstance(item, DgScheduleMetadata)]
        sensors = [item for item in definitions if isinstance(item, DgSensorMetadata)]

        if len(definitions) == 0:
            click.echo("No definitions are defined for this project.")

        console = Console()
        if assets:
            click.echo("Assets")
            table = Table(border_style="dim")
            table.add_column("Key")
            table.add_column("Group")
            table.add_column("Deps")
            table.add_column("Kinds")
            table.add_column("Description")

            for asset in sorted(assets, key=lambda x: x.key):
                table.add_row(
                    asset.key,
                    asset.group,
                    "\n".join(asset.deps),
                    "\n".join(asset.kinds),
                    asset.description,
                )
            console.print(table)
            click.echo("")

        if jobs:
            click.echo("Jobs")
            table = Table(border_style="dim")
            table.add_column("Name")

            for schedule in schedules:
                table.add_row(schedule.name)
            console.print(table)
            click.echo("")

        if schedules:
            click.echo("Schedules")
            table = Table(border_style="dim")
            table.add_column("Name")
            table.add_column("Cron schedule")

            for schedule in schedules:
                table.add_row(schedule.name, schedule.cron_schedule)
            console.print(table)
            click.echo("")

        if sensors:
            click.echo("Sensors")
            table = Table(border_style="dim")
            table.add_column("Name")

            for schedule in schedules:
                table.add_row(schedule.name)
            console.print(table)


def _resolve_definition(item: dict[str, Any]) -> DgDefinitionMetadata:
    if item["type"] == "asset":
        return DgAssetMetadata(**item)
    elif item["type"] == "job":
        return DgJobMetadata(**item)
    elif item["type"] == "schedule":
        return DgScheduleMetadata(**item)
    elif item["type"] == "sensor":
        return DgSensorMetadata(**item)
    else:
        raise DgError(f"Unexpected item type: {item['type']}")



import json
import os
from collections.abc import Mapping
from pathlib import Path
from typing import Any, Optional

import click
import tomlkit
import tomlkit.items

from dagster_dg.component import RemoteComponentRegistry
from dagster_dg.config import (
    DgRawWorkspaceConfig,
    DgWorkspaceScaffoldProjectOptions,
    discover_workspace_root,
)
from dagster_dg.context import DgContext
from dagster_dg.utils import (
    exit_with_error,
    get_toml_node,
    has_toml_node,
    modify_toml,
    scaffold_subtree,
    set_toml_node,
)

# ########################
# ##### WORKSPACE
# ########################


def scaffold_workspace(
    workspace_name: str,
    workspace_config: Optional[DgRawWorkspaceConfig] = None,
) -> Path:
    # Can't create a workspace that is a child of another workspace
    new_workspace_path = Path.cwd() / workspace_name
    existing_workspace_path = discover_workspace_root(new_workspace_path)
    if existing_workspace_path:
        exit_with_error(
            f"Workspace already exists at {existing_workspace_path}.  Run `dg scaffold project` to add a new project to that workspace."
        )
    elif new_workspace_path.exists():
        exit_with_error(f"Folder already exists at {new_workspace_path}.")

    scaffold_subtree(
        path=new_workspace_path,
        name_placeholder="WORKSPACE_NAME_PLACEHOLDER",
        templates_path=os.path.join(
            os.path.dirname(__file__), "templates", "WORKSPACE_NAME_PLACEHOLDER"
        ),
        project_name=workspace_name,
    )

    if workspace_config is not None:
        with modify_toml(new_workspace_path / "pyproject.toml") as toml:
            for k, v in workspace_config.items():
                # Ignore empty collections and None, but not False
                if v != {} and v != [] and v is not None:
                    set_toml_node(toml, ("tool", "dg", "workspace", k), v)

    click.echo(f"Scaffolded files for Dagster workspace at {new_workspace_path}.")
    return new_workspace_path


# ########################
# ##### PROJECT
# ########################


def scaffold_project(
    path: Path,
    dg_context: DgContext,
    use_editable_dagster: Optional[str],
    skip_venv: bool = False,
    populate_cache: bool = True,
) -> None:
    click.echo(f"Creating a Dagster project at {path}.")

    cli_options = DgWorkspaceScaffoldProjectOptions.get_raw_from_cli(use_editable_dagster)
    workspace_options = (
        dg_context.config.workspace.scaffold_project_options
        if dg_context.config.workspace
        else None
    )

    final_use_editable_dagster = cli_options.get(
        "use_editable_dagster",
        workspace_options.use_editable_dagster if workspace_options else None,
    )

    if final_use_editable_dagster:
        editable_dagster_root = (
            _get_editable_dagster_from_env()
            if final_use_editable_dagster is True
            else final_use_editable_dagster
        )
        deps = EDITABLE_DAGSTER_DEPENDENCIES
        dev_deps = EDITABLE_DAGSTER_DEV_DEPENDENCIES
        sources = _gather_dagster_packages(Path(editable_dagster_root))
    else:
        editable_dagster_root = None
        deps = PYPI_DAGSTER_DEPENDENCIES
        dev_deps = PYPI_DAGSTER_DEV_DEPENDENCIES
        sources = []

    dependencies_str = _get_pyproject_toml_dependencies(deps)
    dev_dependencies_str = _get_pyproject_toml_dev_dependencies(dev_deps)
    uv_sources_str = _get_pyproject_toml_uv_sources(sources) if sources else ""

    scaffold_subtree(
        path=path,
        name_placeholder="PROJECT_NAME_PLACEHOLDER",
        templates_path=os.path.join(
            os.path.dirname(__file__), "templates", "PROJECT_NAME_PLACEHOLDER"
        ),
        dependencies=dependencies_str,
        dev_dependencies=dev_dependencies_str,
        uv_sources=uv_sources_str,
    )
    click.echo(f"Scaffolded files for Dagster project at {path}.")

    # Build the venv
    cl_dg_context = dg_context.with_root_path(path)
    if cl_dg_context.use_dg_managed_environment and not skip_venv:
        cl_dg_context.ensure_uv_lock()
        if populate_cache:
            RemoteComponentRegistry.from_dg_context(cl_dg_context)  # Populate the cache

    # Update pyproject.toml
    if cl_dg_context.is_workspace:
        entry = {
            "path": str(cl_dg_context.root_path.relative_to(cl_dg_context.workspace_root_path)),
        }

        with modify_toml(dg_context.pyproject_toml_path) as toml:
            if not has_toml_node(toml, ("tool", "dg", "workspace", "projects")):
                projects = tomlkit.aot()
                set_toml_node(toml, ("tool", "dg", "workspace", "projects"), projects)
                item = tomlkit.table()
            else:
                projects = get_toml_node(
                    toml,
                    ("tool", "dg", "workspace", "projects"),
                    (tomlkit.items.AoT, tomlkit.items.Array),
                )
                if isinstance(projects, tomlkit.items.Array):
                    item = tomlkit.inline_table()
                else:
                    item = tomlkit.table()
            for key, value in entry.items():
                item[key] = value
            projects.append(item)


# Despite the fact that editable dependencies are resolved through tool.uv.sources, we need to set
# the dependencies themselves differently depending on whether we are using editable dagster or
# not. This is because `tool.uv.sources` only seems to apply to direct dependencies of the package,
# so any 2+-order Dagster dependency of our package needs to be listed as a direct dependency in the
# editable case. See: https://github.com/astral-sh/uv/issues/9446
EDITABLE_DAGSTER_DEPENDENCIES = (
    "dagster",
    "dagster-pipes",
    "dagster-shared",
    "dagster-components",
    "dagster-test[components]",  # we include dagster-test for testing purposes
)
EDITABLE_DAGSTER_DEV_DEPENDENCIES = ("dagster-webserver", "dagster-graphql")
PYPI_DAGSTER_DEPENDENCIES = ("dagster-components",)
PYPI_DAGSTER_DEV_DEPENDENCIES = ("dagster-webserver",)


def _get_editable_dagster_from_env() -> str:
    if not os.environ.get("DAGSTER_GIT_REPO_DIR"):
        exit_with_error(
            "The `--use-editable-dagster` option "
            "requires the `DAGSTER_GIT_REPO_DIR` environment variable to be set."
        )
    return os.environ["DAGSTER_GIT_REPO_DIR"]


def _get_pyproject_toml_dependencies(deps: tuple[str, ...]) -> str:
    return "\n".join(
        [
            "dependencies = [",
            *[f'    "{dep}",' for dep in deps],
            "]",
        ]
    )


def _get_pyproject_toml_dev_dependencies(deps: tuple[str, ...]) -> str:
    return "\n".join(
        [
            "dev = [",
            *[f'    "{dep}",' for dep in deps],
            "]",
        ]
    )


def _get_pyproject_toml_uv_sources(lib_paths: list[Path]) -> str:
    lib_lines = [f"{path.name} = {{ path = '{path}', editable = true }}" for path in lib_paths]
    return "\n".join(
        [
            "[tool.uv.sources]",
            *lib_lines,
        ]
    )


def _gather_dagster_packages(editable_dagster_root: Path) -> list[Path]:
    return [
        p.parent
        for p in (
            *editable_dagster_root.glob("python_modules/dagster*/setup.py"),
            *editable_dagster_root.glob("python_modules/libraries/dagster*/setup.py"),
        )
    ]


# ########################
# ##### COMPONENT TYPE
# ########################


def scaffold_component_type(dg_context: DgContext, class_name: str, module_name: str) -> None:
    root_path = Path(dg_context.default_component_library_path)
    click.echo(f"Creating a Dagster component type at {root_path}/{module_name}.py.")

    scaffold_subtree(
        path=root_path,
        name_placeholder="COMPONENT_TYPE_NAME_PLACEHOLDER",
        templates_path=str(Path(__file__).parent / "templates" / "COMPONENT_TYPE"),
        project_name=module_name,
        name=class_name,
    )

    with open(root_path / "__init__.py", "a") as f:
        f.write(
            f"from {dg_context.default_component_library_module_name}.{module_name} import {class_name}\n"
        )

    click.echo(f"Scaffolded files for Dagster component type at {root_path}/{module_name}.py.")


# ########################
# ##### COMPONENT INSTANCE
# ########################


def scaffold_component_instance(
    path: Path,
    component_type: str,
    scaffold_params: Optional[Mapping[str, Any]],
    dg_context: "DgContext",
) -> None:
    click.echo(f"Creating a Dagster component instance folder at {path}.")
    os.makedirs(path, exist_ok=True)
    scaffold_command = [
        "scaffold",
        "component",
        component_type,
        str(path),
        *(["--json-params", json.dumps(scaffold_params)] if scaffold_params else []),
    ]
    dg_context.external_components_command(scaffold_command)



import json
import os
import shutil
import subprocess
import zipfile
from pathlib import Path

import click

from dagster_dg.utils import is_macos, is_windows


def get_default_extension_dir() -> Path:
    if is_windows():
        return Path.home() / "AppData" / "dg" / "vscode"
    elif is_macos():
        return Path.home() / "Library" / "Application Support" / "dg" / "vscode"
    else:
        return Path.home() / ".local" / "share" / "dg" / "vscode"


def has_editor_cli_command(executable_name: str) -> bool:
    return bool(shutil.which(executable_name))


def run_editor_cli_command(executable_name: str, args: list[str]) -> bytes:
    return subprocess.check_output([executable_name] + args)


def recommend_yaml_extension(executable_name: str) -> None:
    if not has_editor_cli_command(executable_name):
        raise click.ClickException(
            f"Could not find `{executable_name}` executable in PATH. In order to use the dagster-components-schema extension, "
            "please install the redhat.vscode-yaml extension manually."
        )

    extensions = (
        run_editor_cli_command(executable_name, ["--list-extensions"]).decode("utf-8").split("\n")
    )
    if "redhat.vscode-yaml" in extensions:
        click.echo("redhat.vscode-yaml extension is already installed.")
    else:
        if click.confirm(
            "The redhat.vscode-yaml extension is not installed. Would you like to install it now?"
        ):
            run_editor_cli_command(executable_name, ["--install-extension", "redhat.vscode-yaml"])


def install_or_update_yaml_schema_extension(
    executable_name: str, yaml_dir: Path, schema_path: Path
) -> None:
    """Builds a VS Code extension which associates the built JSON schema files with YAML
    files in the provided directory, provided that the user has the Red Hat YAML extension
    already installed.
    """
    extension_working_dir = get_default_extension_dir() / "dagster-components-schema"
    extension_package_json_path = extension_working_dir / "package.json"

    template_package_json_path = Path(__file__).parent / "vscode_extension_package.json"
    template_package_json = json.loads(template_package_json_path.read_text())
    template_package_json["contributes"]["yamlValidation"] = [
        {"fileMatch": f"{yaml_dir}/**/*.y*ml", "url": f"{schema_path}"}
    ]

    extension_working_dir.mkdir(parents=True, exist_ok=True)

    # Merge with existing yamlValidation entries, so we can provide schema completions for many
    # projects.
    if extension_package_json_path.exists():
        existing_package_json = json.loads(extension_package_json_path.read_text())
        existing_yaml_validation = existing_package_json["contributes"].get("yamlValidation")
        if existing_yaml_validation:
            template_package_json["contributes"]["yamlValidation"].extend(
                entry
                for entry in existing_yaml_validation
                if entry["fileMatch"] != f"{yaml_dir}/**/*.y*ml"
            )

    extension_package_json_path.write_text(json.dumps(template_package_json, indent=2))
    click.echo(f"Set up package.json for VS Code extension in {extension_package_json_path}")

    # Local VS Code extensions must be packaged into a vsix file, which under the hood is just a zip file
    # with a special extension.
    extension_zip_path = extension_working_dir / "dagster-components-schema.vsix"
    with zipfile.ZipFile(extension_zip_path, "w") as z:
        z.write(extension_package_json_path, "extension/package.json")

    click.echo(f"Packaged extension to {extension_zip_path}")

    try:
        run_editor_cli_command(
            executable_name, ["--uninstall-extension", "dagster.dagster-components-schema"]
        )
    except subprocess.CalledProcessError:
        click.echo("No existing dagster.dagster-components-schema extension to uninstall.")
    run_editor_cli_command(
        executable_name, ["--install-extension", os.fspath(extension_zip_path.resolve())]
    )
    click.echo("Successfully installed Dagster Components schema extension.")



## Lifted from dagster._utils.yaml_utils
from collections.abc import Mapping, Sequence
from typing import Any, cast

import yaml

from dagster_dg.yaml_utils.source_position import (
    KeyPathSegment,
    LineCol,
    SourcePosition,
    SourcePositionTree,
    ValueAndSourcePositionTree,
)

# Removing this implicit loader stops YAML-loading from returning datetime objects
YAML_TIMESTAMP_TAG = "tag:yaml.org,2002:timestamp"


class _CanRemoveImplicitResolver:
    # Adds a "remove_implicit_resolver" method that can be used to selectively
    # disable default PyYAML resolvers

    @classmethod
    def remove_implicit_resolver(cls, tag):
        # See https://github.com/yaml/pyyaml/blob/master/lib/yaml/resolver.py#L26 for inspiration
        if "yaml_implicit_resolvers" not in cls.__dict__:
            implicit_resolvers = {}
            for key in cls.yaml_implicit_resolvers:
                implicit_resolvers[key] = cls.yaml_implicit_resolvers[key][:]
            cls.yaml_implicit_resolvers = implicit_resolvers

        for ch, mappings in cls.yaml_implicit_resolvers.items():
            cls.yaml_implicit_resolvers[ch] = [
                (existing_tag, regexp) for existing_tag, regexp in mappings if existing_tag != tag
            ]


def parse_yaml_with_source_positions(
    src: str,
    filename: str = "<string>",
    implicits_to_remove: Sequence[str] = [YAML_TIMESTAMP_TAG],
) -> ValueAndSourcePositionTree:
    """Parse YAML source with source position information.
    This function takes a YAML source string and an optional filename, and returns a
    `ValueAndSourcePositionTree` object. The `ValueAndSourcePositionTree` contains the
    parsed YAML value and a `SourcePositionTree` that holds the source position information for
    the YAML structure.

    The function uses a custom YAML loader (`SourcePositionLoader`) that tracks the source position
    information for each YAML node. It also allows for the removal of specific implicit resolvers,
    such as the default timestamp resolver.

    Args:
        src (str): The YAML source string to be parsed.
        filename (str): The filename associated with the YAML source, used for error reporting.
            Defaults to "<string>" if not provided.
        implicits_to_remove (list[str]): A list of YAML implicit resolvers to remove. Defaults to
            removing the timestamp resolver.

    Returns:
        ValueAndSourcePositionTree: An object containing the parsed YAML value and the corresponding
            source position information.
    """

    class SourcePositionLoader(yaml.SafeLoader, _CanRemoveImplicitResolver):
        def construct_object(self, node: Any, deep: Any = False) -> ValueAndSourcePositionTree:
            """Returns a ValueAndSourcePositionTree object where:
            - Its value is what the regular yaml.SafeLoader returns
            - Its source_position_tree is a tree where each node corresponds to a node in the YAML
              tree.
            """
            source_position = SourcePosition(
                filename=filename,
                start=LineCol(line=node.start_mark.line + 1, col=node.start_mark.column + 1),
                end=LineCol(line=node.end_mark.line + 1, col=node.end_mark.column + 1),
            )

            value = super().construct_object(node, True)
            # If value is a collection, then all of its elements will be ValueAndSourcePositionTree
            # instances. We need to do two things:
            # - Strip out the enclosing ValueAndSourcePositionTrees, so we're just returning a
            #   collection of the raw values
            # - Assemble the source position subtrees into a super-tree

            if isinstance(value, dict):
                dict_with_raw_values: dict[Any, Any] = {}
                child_trees: dict[KeyPathSegment, SourcePositionTree] = {}
                for k, v in cast(
                    Mapping[ValueAndSourcePositionTree, ValueAndSourcePositionTree], value
                ).items():
                    dict_with_raw_values[k.value] = v.value
                    child_trees[k.value] = SourcePositionTree(
                        position=k.source_position_tree.position,
                        children=v.source_position_tree.children,
                    )
                return ValueAndSourcePositionTree(
                    dict_with_raw_values,
                    SourcePositionTree(position=source_position, children=child_trees),
                )
            if isinstance(value, list):
                list_with_raw_values: list[Any] = []
                child_trees: dict[KeyPathSegment, SourcePositionTree] = {}
                for i, v in enumerate(cast(Sequence[ValueAndSourcePositionTree], value)):
                    list_with_raw_values.append(v.value)
                    child_trees[i] = v.source_position_tree
                return ValueAndSourcePositionTree(
                    list_with_raw_values,
                    SourcePositionTree(position=source_position, children=child_trees),
                )

            return ValueAndSourcePositionTree(
                value,
                SourcePositionTree(position=source_position, children={}),
            )

    for implicit_to_remove in implicits_to_remove:
        SourcePositionLoader.remove_implicit_resolver(implicit_to_remove)

    try:
        value_and_tree_node = yaml.load(src, Loader=SourcePositionLoader)
    except yaml.MarkedYAMLError as e:
        if e.context_mark is not None:
            e.context_mark.name = filename
        if e.problem_mark is not None:
            e.problem_mark.name = filename
        raise e

    return value_and_tree_node
